"""End-to-End (E2E) tests for this library."""

from __future__ import annotations

from pathlib import Path

from pytest import mark
from syrupy.assertion import SnapshotAssertion as Snapshot

import magodo

from .data import get_all_todo_paths


params = mark.parametrize


@params("todo_file", get_all_todo_paths())
def test_group_From_file(snapshot: Snapshot, todo_file: Path) -> None:
    """Test the TodoGroup.from_file() function."""
    todo_group = magodo.TodoGroup.from_path(todo_file)
    assert [todo.__dict__ for todo in todo_group] == snapshot



"""Contains the Todo class definition."""

# pylint: disable=format-string-without-interpolation

from __future__ import annotations

from dataclasses import dataclass
import datetime as dt
import re
from typing import Final, List, Optional, Tuple, cast

from eris import ErisError, Err, Ok, Result

from ._dates import RE_DATE, to_date
from .types import Metadata, Priority


re_todo_fmt = r"""
(?P<x>x[ ]+)?                        # optional 'x'
{{0}}                                # priority
(?:
    (?:(?P<done_date>{0})[ ]+)?      # optional date of completion
    (?:(?P<create_date>{0})[ ]+)     # optional date of creation
)?
(?P<desc>[A-Za-z0-9+@].*)            # description
""".format(
    RE_DATE
).format
re_opt_fmt = r"(?:{})?".format

CONTEXT_PREFIX: Final = "@"
DEFAULT_PRIORITY: Final[Priority] = "N"
PROJECT_PREFIX: Final = "+"
PUNCTUATION: Final = ",.?!;"
RE_PRIORITY: Final = r"\((?P<priority>[A-Z])\)[ ]+"
RE_OPTIONAL_PRIORITY: Final = re_opt_fmt(RE_PRIORITY)


@dataclass(frozen=True)
class Todo:
    """Represents a single task in a todo list."""

    desc: str

    contexts: Tuple[str, ...] = ()
    create_date: Optional[dt.date] = None
    done_date: Optional[dt.date] = None
    marked_done: bool = False
    metadata: Optional[Metadata] = None
    priority: Priority = DEFAULT_PRIORITY
    projects: Tuple[str, ...] = ()

    @classmethod
    def from_line(
        cls, line: str, *, strict: bool = False
    ) -> Result[Todo, ErisError]:
        """Contructs a Todo object from a string (usually a line in a file).

        Args:
            line: The line to use to construct our new Todo object.
            strict: If this option is set, require that valid todo lines
              either explicitly specify a priority OR have been marked done /
              not done with an "x " or "o " at the start of the line,
              respectively. NOTE: The special "o " prefix is treated as
              short-hand for the next letter after the default priority letter
              and is ONLY allowed when this option is set.
        """
        line = line.strip()

        if (
            strict
            and line
            and line.startswith("x:")
            and line[2:6].isnumeric()
            and line[6].isspace()
        ):
            xword, *rest = line.split(" ")
            dtime = xword[2:]
            new_line_words = ["x"] + rest + ["dtime:" + dtime]
            line = " ".join(new_line_words)

        if strict and line.startswith("o "):
            o_priority = (
                DEFAULT_PRIORITY
                if DEFAULT_PRIORITY == "Z"
                else chr(ord(DEFAULT_PRIORITY) + 1)
            )
            line = f"({o_priority})" + line[1:]

        if strict and not line.startswith("x "):
            RE_TODO = re_todo_fmt(RE_PRIORITY)
        else:
            RE_TODO = re_todo_fmt(RE_OPTIONAL_PRIORITY)

        re_todo_match = re.match(RE_TODO, line, re.VERBOSE)
        if re_todo_match is None:
            return Err(
                f"The provided string ({line!r}) does not appear to properly"
                " adhere to the todo.txt format. See"
                " https://github.com/todotxt/todo.txt for the specification. "
                f" |  require_priority={strict}"
            )

        marked_done: bool = False
        if re_todo_match.group("x"):
            marked_done = True

        priority: Priority = DEFAULT_PRIORITY
        if grp := re_todo_match.group("priority"):
            priority = cast(Priority, grp)

        create_date: Optional[dt.date] = None
        if grp := re_todo_match.group("create_date"):
            create_date = to_date(grp)

        done_date: Optional[dt.date] = None
        if grp := re_todo_match.group("done_date"):
            done_date = to_date(grp)

        desc = re_todo_match.group("desc")
        all_words = desc.split(" ")

        project_list: List[str] = []
        context_list: List[str] = []
        for some_list, prefix in [
            (project_list, PROJECT_PREFIX),
            (context_list, CONTEXT_PREFIX),
        ]:
            for word in all_words:
                if (
                    word.startswith(prefix)
                    and not word.startswith(prefix + prefix)
                    and len(prefix) < len(word)
                ):
                    value = word[len(prefix) :]
                    value = _clean_value(value)
                    some_list.append(value)

        projects = tuple(project_list)
        contexts = tuple(context_list)

        metadata: Optional[Metadata] = None
        mdata: Metadata = {}
        for word in all_words:
            kv = word.split(":", maxsplit=1)
            if len(kv) == 2 and kv[1] and not kv[1].startswith(":"):
                key, value = kv
                value = _clean_value(value)

                if key in mdata:
                    value_list = mdata[key]
                    if not isinstance(value_list, list):
                        value_list = [value_list]
                        mdata[key] = value_list

                    value_list.append(value)
                else:
                    mdata[key] = value

        if mdata:
            metadata = mdata

        todo = cls(
            contexts=contexts,
            create_date=create_date,
            desc=desc,
            done_date=done_date,
            marked_done=marked_done,
            metadata=metadata,
            priority=priority,
            projects=projects,
        )
        return Ok(todo)


def _clean_value(word: str) -> str:
    """Cleanup context, metadata, or project value.

    Makes the following changes to `word`:

      - Strips any punctuation from the right-side of `word`.
      - Removes any possesive apostrophe at the end of `word`.

    NOTE: Will not strip punctuation if `word` is composed ONLY of punctuation
      characters.
    """
    result = word.rstrip(PUNCTUATION)
    if not result:
        result = word

    result = result.split("'", maxsplit=1)[0]
    return result



"""Custom types used by this library."""

from __future__ import annotations

from typing import Dict, List, Literal, Union


# Type of the Todo.metadata attribute.
Metadata = Dict[str, Union[str, List[str]]]
# A todo item's priority is always a capital letter.
Priority = Literal[
    "A",
    "B",
    "C",
    "D",
    "E",
    "F",
    "G",
    "H",
    "I",
    "J",
    "K",
    "L",
    "M",
    "N",
    "O",
    "P",
    "Q",
    "R",
    "S",
    "T",
    "U",
    "V",
    "W",
    "X",
    "Y",
    "Z",
]



"""A Python library for working with the todo.txt format."""

import logging as _logging

from . import types
from ._group import TodoGroup
from ._todo import DEFAULT_PRIORITY, Todo


__all__ = ["DEFAULT_PRIORITY", "Todo", "TodoGroup", "types"]

__author__ = "Bryan M Bugyi"
__email__ = "bryanbugyi34@gmail.com"
__version__ = "0.4.1"

_logging.getLogger(__name__).addHandler(_logging.NullHandler())



"""Datetime utilities used throughout this library."""

from __future__ import annotations

import datetime as dt
from typing import Final


DATE_FMT: Final = "%Y-%m-%d"
RE_DATE: Final = r"[1-9][0-9][0-9][0-9]-[0-1][0-9]-[0-3][0-9]"


def to_date(yyyymmdd: str) -> dt.date:
    """Helper function for constructing a date object."""
    return dt.datetime.strptime(yyyymmdd, DATE_FMT).date()



"""Tests for the magodo package."""

from __future__ import annotations

from pytest import mark

from magodo import Todo
from magodo._dates import to_date


params = mark.parametrize


@params(
    "line,strict,expected",
    [
        ("no priority todo", False, Todo(desc="no priority todo")),
        ("(N) basic todo", False, Todo(desc="basic todo")),
        (
            "(N) 2022-01-10 todo with create date",
            True,
            Todo(
                desc="todo with create date",
                create_date=to_date("2022-01-10"),
            ),
        ),
        (
            "(N) 2022-03-04 2022-01-10 todo with done date",
            True,
            Todo(
                desc="todo with done date",
                create_date=to_date("2022-01-10"),
                done_date=to_date("2022-03-04"),
            ),
        ),
        (
            "x (N) 2022-01-10 done todo",
            True,
            Todo(
                desc="done todo",
                create_date=to_date("2022-01-10"),
                marked_done=True,
            ),
        ),
        (
            "x (A) todo with priority",
            True,
            Todo(
                desc="todo with priority",
                marked_done=True,
                priority="A",
            ),
        ),
        (
            "(N) todo for +some +project",
            True,
            Todo(
                desc="todo for +some +project",
                projects=("some", "project"),
            ),
        ),
        (
            "(N) todo for +some +project and a @context.",
            True,
            Todo(
                desc="todo for +some +project and a @context.",
                projects=("some", "project"),
                contexts=("context",),
            ),
        ),
        (
            "(N) todo with +some meta:data and a @context due:2022-12-31",
            True,
            Todo(
                desc="todo with +some meta:data and a @context due:2022-12-31",
                projects=("some",),
                contexts=("context",),
                metadata={"meta": "data", "due": "2022-12-31"},
            ),
        ),
        (
            "(N) todo with some dep:123, another dep:456, and a 3rd dep:789..."
            " foo:bar @crazy",
            True,
            Todo(
                desc=(
                    "todo with some dep:123, another dep:456, and a 3rd"
                    " dep:789... foo:bar @crazy"
                ),
                contexts=("crazy",),
                metadata={"dep": ["123", "456", "789"], "foo": "bar"},
            ),
        ),
        (
            "x:1030 Shorthand 'dtime' syntax.",
            True,
            Todo(
                desc="Shorthand 'dtime' syntax. dtime:1030",
                metadata={"dtime": "1030"},
                marked_done=True,
            ),
        ),
        (
            "x:1030 2022-01-01 Bug with shorthand 'dtime' syntax.",
            True,
            Todo(
                desc="Bug with shorthand 'dtime' syntax. dtime:1030",
                metadata={"dtime": "1030"},
                marked_done=True,
                create_date=to_date("2022-01-01"),
            ),
        ),
    ],
)
def test_todo(line: str, strict: bool, expected: Todo) -> None:
    """Test the Todo type."""
    actual = Todo.from_line(line, strict=strict)
    assert actual.unwrap() == expected



"""Contains the TodoGroup class definition."""

from __future__ import annotations

from logging import Logger
from pathlib import Path
from typing import Iterable, Iterator, List

from eris import Err
from typist import PathLike

from ._todo import Todo


logger = Logger(__name__)


class TodoGroup:
    """Manages a group of Todo objects."""

    def __init__(self, todos: Iterable[Todo]) -> None:
        self._todos = list(todos)

    def __repr__(self) -> str:  # noqa: D105
        return f"{self.__class__.__name__}(todos={self._todos})"

    def __iter__(self) -> Iterator[Todo]:
        """Yields the Todo objects that belong to this group."""
        yield from self._todos

    @classmethod
    def from_path(cls, path_like: PathLike) -> TodoGroup:
        """Reads all todo lines from a given file or directory (recursively).

        Pre-conditions:
            * `path_like` exists and is either a file or directory.
        """
        path = Path(path_like)

        assert path.exists(), f"The provided path does not exist: {path}"

        todos: List[Todo] = []
        if path.is_file():
            logger.debug(
                "Attempting to load todos from text file: file=%r", str(path)
            )

            for line in path.read_text().split("\n"):
                line = line.strip()
                todo_result = Todo.from_line(line, strict=True)
                if isinstance(todo_result, Err):
                    continue

                todo = todo_result.ok()
                todos.append(todo)
                logger.debug("New todo loaded: todo=%r", todo)
        else:
            assert path.is_dir(), (
                "The provided path exists but is neither a file nor a"
                f" directory: {path}"
            )

            logger.debug("Loading from directory: dir=%r", str(path))
            for other_path in path.glob("*"):
                if other_path.is_file() and other_path.suffix != ".txt":
                    continue

                other_todo_group = TodoGroup.from_path(other_path)
                todos.extend(other_todo_group)

        return cls(todos)
